/***************************************************************************
  This is a library for the BMP085 pressure sensor

  Designed specifically to work with the Adafruit BMP085 or BMP180 Breakout 
  ----> http://www.adafruit.com/products/391
  ----> http://www.adafruit.com/products/1603
 
  These displays use I2C to communicate, 2 pins are required to interface.

  Adafruit invests time and resources providing this open source code,
  please support Adafruit andopen-source hardware by purchasing products
  from Adafruit!

  Written by Kevin Townsend for Adafruit Industries.  
  BSD license, all text above must be included in any redistribution
 ***************************************************************************/
#include <wiringPiI2C.h>

//#include <Wire.h>
#include <math.h>
#include <limits.h>
#include <unistd.h>
#include <iostream>

#include "include/Adafruit_BMP085_U.h"

static bmp085_calib_data _bmp085_coeffs;   // Last read accelerometer data will be available here
static unsigned int           _bmp085Mode;
volatile int fd; //linux file handle used by wiringPiI2C

#define BMP085_USE_DATASHEET_VALS (0) /* Set to 1 for sanity check */
#define DEBUG (0) //prints lots of extra stuff

using namespace std;

/***************************************************************************
 PRIVATE FUNCTIONS
 ***************************************************************************/

/**************************************************************************/
/*!
    @brief  Writes an 8 bit value over I2C
*/
/**************************************************************************/
static void writeCommand(unsigned char reg, unsigned char value)
{
  /*
  Wire.beginTransmission((unsigned int)BMP085_ADDRESS);
  #if ARDUINO >= 100
    Wire.write((unsigned int)reg);
    Wire.write((unsigned int)value);
  #else
    Wire.send(reg);
    Wire.send(value);
  #endif
  Wire.endTransmission();
  */
  // •int wiringPiI2CWriteReg8 (int fd, int reg, int data) ;
  wiringPiI2CWriteReg8 (fd, (unsigned char)reg, (unsigned char)value) ;
}

/**************************************************************************/
/*!
    @brief  Reads an 8 bit value over I2C
*/
/**************************************************************************/
static void read8(unsigned char reg, unsigned short *value)
{
/*
  Wire.beginTransmission((unsigned int)BMP085_ADDRESS);
  #if ARDUINO >= 100
    Wire.write((unsigned int)reg);
  #else
    Wire.send(reg);
  #endif
  Wire.endTransmission();
  Wire.requestFrom((unsigned int)BMP085_ADDRESS, (byte)1);
  #if ARDUINO >= 100
    *value = Wire.read();
  #else
    *value = Wire.receive();
  #endif  
  Wire.endTransmission();
  */
  //*value = int wiringPiI2CReadReg8 (int fd, int reg);
  *value = wiringPiI2CReadReg8 (fd, reg);
}

/**************************************************************************/
/*!
    @brief  Reads a 16 bit value over I2C
*/
/**************************************************************************/
static void read16(unsigned char reg, unsigned short *value)
{
  unsigned short hi = wiringPiI2CReadReg8 (fd, reg);
  unsigned short lo = wiringPiI2CReadReg8 (fd, reg+1);
  *value = (unsigned short)((hi << 8) + lo);
}

/**************************************************************************/
/*!
    @brief  Reads a signed 16 bit value over I2C
*/
/**************************************************************************/
static void readS16(unsigned char reg, short *value)
{
  short hi = wiringPiI2CReadReg8 (fd, reg);
  unsigned short lo = wiringPiI2CReadReg8 (fd, reg+1);
  *value = (signed short)((hi << 8) + lo);
}

/**************************************************************************/
/*!
    @brief  Reads the factory-set coefficients
*/
/**************************************************************************/
static void readCoefficients(void)
{
  #if BMP085_USE_DATASHEET_VALS
    _bmp085_coeffs.ac1 = 408;
    _bmp085_coeffs.ac2 = -72;
    _bmp085_coeffs.ac3 = -14383;
    _bmp085_coeffs.ac4 = 32741;
    _bmp085_coeffs.ac5 = 32757;
    _bmp085_coeffs.ac6 = 23153;
    _bmp085_coeffs.b1  = 6190;
    _bmp085_coeffs.b2  = 4;
    _bmp085_coeffs.mb  = -32768;
    _bmp085_coeffs.mc  = -8711;
    _bmp085_coeffs.md  = 2868;
    _bmp085Mode        = 0;
  #else
    readS16(BMP085_REGISTER_CAL_AC1, &_bmp085_coeffs.ac1);
    readS16(BMP085_REGISTER_CAL_AC2, &_bmp085_coeffs.ac2);
    readS16(BMP085_REGISTER_CAL_AC3, &_bmp085_coeffs.ac3);
    read16(BMP085_REGISTER_CAL_AC4, &_bmp085_coeffs.ac4);
    read16(BMP085_REGISTER_CAL_AC5, &_bmp085_coeffs.ac5);
    read16(BMP085_REGISTER_CAL_AC6, &_bmp085_coeffs.ac6);
    readS16(BMP085_REGISTER_CAL_B1, &_bmp085_coeffs.b1);
    readS16(BMP085_REGISTER_CAL_B2, &_bmp085_coeffs.b2);
    readS16(BMP085_REGISTER_CAL_MB, &_bmp085_coeffs.mb);
    readS16(BMP085_REGISTER_CAL_MC, &_bmp085_coeffs.mc);
    readS16(BMP085_REGISTER_CAL_MD, &_bmp085_coeffs.md);
  #endif
  
  #if DEBUG
	cout << _bmp085_coeffs.ac1
	<< endl << _bmp085_coeffs.ac2    
	<< endl << _bmp085_coeffs.ac3
	<< endl << _bmp085_coeffs.ac4    
	<< endl << _bmp085_coeffs.ac5
	<< endl << _bmp085_coeffs.ac6
	<< endl << _bmp085_coeffs.b1
	<< endl << _bmp085_coeffs.b2
	<< endl << _bmp085_coeffs.mb
	<< endl << _bmp085_coeffs.mc
	<< endl << _bmp085_coeffs.md << endl;
  #endif
}

/**************************************************************************/
/*!

*/
/**************************************************************************/
static void readRawTemperature(int *temperature)
{
  #if BMP085_USE_DATASHEET_VALS
    *temperature = 27898;
  #else
    unsigned short t;
    writeCommand(BMP085_REGISTER_CONTROL, BMP085_REGISTER_READTEMPCMD);
    usleep(5000);
    read16(BMP085_REGISTER_TEMPDATA, &t);
    *temperature = (unsigned int)t;
  #endif
  
  #if DEBUG
	cout << "Raw Temp: " << t << endl;
  #endif
}

/**************************************************************************/
/*!

*/
/**************************************************************************/
static void readRawPressure(int *pressure)
{
  #if BMP085_USE_DATASHEET_VALS
    *pressure = 23843;
  #else
    unsigned short  msb, lsb, xlsb;
    int  raw;

    writeCommand(BMP085_REGISTER_CONTROL, BMP085_REGISTER_READPRESSURECMD + (_bmp085Mode << 6));
    switch(_bmp085Mode)
    {
      case BMP085_MODE_ULTRALOWPOWER:
        usleep(5000);
        break;
      case BMP085_MODE_STANDARD:
        usleep(8000);
        break;
      case BMP085_MODE_HIGHRES:
        usleep(14000);
        break;
      case BMP085_MODE_ULTRAHIGHRES:
      default:
        usleep(26000);
        break;
    }

    read8(BMP085_REGISTER_PRESSUREDATA, &msb);
    read8(BMP085_REGISTER_PRESSUREDATA+1, &lsb);
    read8(BMP085_REGISTER_PRESSUREDATA+2, &xlsb);
	raw = ((msb << 16) + (lsb << 8) + xlsb) >> (8 - _bmp085Mode);
    
    *pressure = raw;
  #endif
  #if DEBUG
	cout 	<< "Raw Pressure: " << raw << endl;
  #endif
}

/***************************************************************************
 CONSTRUCTOR
 ***************************************************************************/
 
/**************************************************************************/
/*!
    @brief  Instantiates a new Adafruit_BMP085_Unified class
*/
/**************************************************************************/
Adafruit_BMP085_Unified::Adafruit_BMP085_Unified(int sensorID) {
  _sensorID = sensorID;
}

/***************************************************************************
 PUBLIC FUNCTIONS
 ***************************************************************************/
 
/**************************************************************************/
/*!
    @brief  Setups the HW
*/
/**************************************************************************/
int Adafruit_BMP085_Unified::begin(bmp085_mode_t mode)
{
  // Enable I2C
  //Wire.begin();
  fd = wiringPiI2CSetup(BMP085_ADDRESS);
  if(fd < 0){
	return fd;
  }

  /* Mode boundary check */
  if ((mode > BMP085_MODE_ULTRAHIGHRES) || (mode < 0))
  {
    mode = BMP085_MODE_ULTRAHIGHRES;
  }

  /* Make sure we have the right device */
  unsigned short id;
  read8(BMP085_REGISTER_CHIPID, &id);
  if(id != 0x55)
  {
    return -100;
  }

  /* Set the mode indicator */
  _bmp085Mode = mode;

  /* Coefficients need to be read once */
  readCoefficients();
  
  return 1;
}

/**************************************************************************/
/*!
    @brief  Gets the compensated pressure level in kPa
*/
/**************************************************************************/
void Adafruit_BMP085_Unified::getPressure(float *pressure)
{
  int  ut = 0, up = 0, compp = 0;
  int  x1, x2, b5, b6, x3, b3, p;
  unsigned int b4, b7;

  /* Get the raw pressure and temperature values */
  readRawTemperature(&ut);
  readRawPressure(&up);

  /* Temperature compensation */
  x1 = (ut - (int)(_bmp085_coeffs.ac6))*((int)(_bmp085_coeffs.ac5)) >> 15;
  x2 = ((int)(_bmp085_coeffs.mc << 11))/(x1+(int)(_bmp085_coeffs.md));
  b5 = x1 + x2;
  
  /* Pressure compensation */
  b6 = b5 - 4000;
  x1 = (_bmp085_coeffs.b2 * ((b6 * b6) >> 12)) >> 11;
  x2 = (_bmp085_coeffs.ac2 * b6) >> 11;
  x3 = x1 + x2;
  b3 = (((((int) _bmp085_coeffs.ac1) * 4 + x3) << _bmp085Mode) + 2) >> 2;
  
  #if DEBUG
	cout 	<< "B6: " << b6 << endl
			<< "X1: " << x1 << endl
			<< "X2: " << x2 << endl
			<< "X3: " << x3 << endl
			<< "B3: " << b3 << endl;
  #endif
  
  x1 = (_bmp085_coeffs.ac3 * b6) >> 13;
  x2 = (_bmp085_coeffs.b1 * ((b6 * b6) >> 12)) >> 16;
  x3 = ((x1 + x2) + 2) >> 2;
  b4 = (_bmp085_coeffs.ac4 * (unsigned int) (x3 + 32768)) >> 15;
  b7 = ((unsigned int) (up - b3) * (50000 >> _bmp085Mode));

  
  #if DEBUG
	cout 	<< "X1: " << x1 << endl
			<< "X2: " << x2 << endl
			<< "X3: " << x3 << endl
			<< "B4: " << b4 << endl
			<< "B7: " << b7 << endl;
  #endif
  
  if (b7 < 0x80000000)
  {
    p = (b7 << 1) / b4;
  }
  else
  {
    p = (b7 / b4) << 1;
  }
  #if DEBUG
	cout 	<< "X1: " << x1 << endl;
  #endif

  x1 = (p >> 8) * (p >> 8);
  x1 = (x1 * 3038) >> 16;
  x2 = (-7357 * p) >> 16;
  
  
  #if DEBUG
	cout 	<< "p: " << p << endl
			<< "X1: " << x1 << endl
			<< "X2: " << x2 << endl;
  #endif
  
  compp = p + ((x1 + x2 + 3791) >> 4);

  
  #if DEBUG
	cout 	<< "Pressure: " << p << endl;
  #endif
  
  /* Assign compensated pressure value */
  *pressure = compp;
}

/**************************************************************************/
/*!
    @brief  Reads the temperatures in degrees Celsius
*/
/**************************************************************************/
void Adafruit_BMP085_Unified::getTemperature(float *temp)
{
  int UT, X1, X2, B5;     // following ds convention
  float t;

  readRawTemperature(&UT);

  #if BMP085_USE_DATASHEET_VALS
    // use datasheet numbers!
    UT = 27898;
    _bmp085_coeffs.ac6 = 23153;
    _bmp085_coeffs.ac5 = 32757;
    _bmp085_coeffs.mc = -8711;
    _bmp085_coeffs.md = 2868;
  #endif

  // step 1
  X1 = ((UT - (unsigned int)_bmp085_coeffs.ac6) * ((unsigned int)_bmp085_coeffs.ac5)) >> 15;
  X2 = ((int)_bmp085_coeffs.mc << 11) / (X1+(int)_bmp085_coeffs.md);
  B5 = X1 + X2;
  t = (B5+8) >> 4;
  t /= 10.0;

  #if DEBUG
	cout 	<< "X1: " << X1 << endl
			<< "X2: " << X2 << endl
			<< "B5: " << B5 << endl;
  #endif
  
  *temp = t;
}

/**************************************************************************/
/*!
    Calculates the altitude (in meters) from the specified atmospheric
    pressure (in hPa), sea-level pressure (in hPa), and temperature (in �C)

    @param  seaLevel      Sea-level pressure in hPa
    @param  atmospheric   Atmospheric pressure in hPa
    @param  temp          Temperature in degrees Celsius
*/
/**************************************************************************/
float Adafruit_BMP085_Unified::pressureToAltitude(float seaLevel, float atmospheric, float temp)
{
  /* Hyposometric formula:                      */
  /*                                            */
  /*     ((P0/P)^(1/5.257) - 1) * (T + 273.15)  */
  /* h = -------------------------------------  */
  /*                   0.0065                   */
  /*                                            */
  /* where: h   = height (in meters)            */
  /*        P0  = sea-level pressure (in hPa)   */
  /*        P   = atmospheric pressure (in hPa) */
  /*        T   = temperature (in degC)         */

  return (((float)pow((seaLevel/atmospheric), 0.190223F) - 1.0F)
         * (temp + 273.15F)) / 0.0065F;
}