/*
* File:   main.cpp
* Author: Anton Volkov, Aleksey Lykov, William Tarpley
* Multithreaded version without hardware (3 separate threads: GPS, INS, PROC)
*  No serial, I2C (mpu6050), or curses libraries; windows.h instead of unistd.h
*
* Created on November 5, 2013, 2:33 PM
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include <iostream>
#include <iomanip>
#include <time.h>
#include <fstream>
#include <sstream>
#include <ncurses.h>

//Serial Libraries:
#include <wiringPi.h>
#include <wiringSerial.h>

//BMP180 Library:
#include "include/Adafruit_BMP085_U.h"

//MPU-9150 Libraries:
#include "include/I2Cdev.h"
#include "include/MPU6050.h"
#include "include/MPU6050_6Axis_MotionApps20.h"
#include "include/helper_3dmath.h"

//For delays in no-HW Windows testing:
// #include <windows.h>

//POSIX 
#include <pthread.h>
#define NUMTHREADS 4 //ins, gps, bar, proc
// #define NUMTHREADS 3

using namespace std;

pthread_mutex_t insfile_mutex; //ins.txt file control mutex
pthread_mutex_t gpsfile_mutex; //gps.txt file control mutex
pthread_mutex_t barfile_mutex; //bar.txt file control mutex
pthread_mutex_t procfile_mutex; //proc.txt file control mutex
pthread_mutex_t procrawfile_mutex; //procraw.txt file control mutex

pthread_t insThread;
pthread_t gpsThread;
pthread_t barThread;
pthread_t procThread;

/* Raw data incl. magnetomter:
MPU6050 mpu;

int16_t ax, ay, az;
int16_t gx, gy, gz;
int16_t mx, my, mz;

The above data can be polled from MPU9150 via: MPU6050::getMotion9
this commands calls getMotion6 (for ax, ay, az, gx, gy, gz) then
reads magnetometer (must wait 12-20ms for proper reading, returns mx, my, mz
code from getMotion9:

    // magnetometer reading
    i2Cdev.writeByte(devAddr, MPU6050_RA_INT_PIN_CFG, 0x02); //set i2c bypass enable pin to true to access magnetometer
    wait_ms(10); // necessary wait >=6ms
    i2Cdev.writeByte(MPU9150_RA_MAG_ADDRESS, 0x0A, 0x01); // enable the magnetometer
    wait_ms(10); // necessary wait >=6ms
    i2Cdev.readBytes(MPU9150_RA_MAG_ADDRESS, MPU9150_RA_MAG_XOUT_L, 6, buffer);
    *mx = (((int16_t)buffer[0]) << 8) | buffer[1];
    *my = (((int16_t)buffer[2]) << 8) | buffer[3];
    *mz = (((int16_t)buffer[4]) << 8) | buffer[5];
*/
///////////////////////////////////////////////////////////////////////////

//output files (text documents accessed by all functions one at a time)
// FILE * procFile;
// FILE * procRawFile;
ofstream procFile;
ofstream procRawFile;

stringstream ssBuffer1, ssBuffer2; //temporary storage of data to be saved to SD card
int resMem = 50; //Megabytes
///////////////////////////////////////////////////////////////////////////

//Thread functions:
void *thread_INS(void*);
void *thread_GPS(void*);
void *thread_BAR(void*);
void *thread_PROC(void*);
//Mag start procedure functions:
void magInit();
void getMag(int16_t* mx, int16_t* my, int16_t* mz);
void readMagData(int16_t* mx, int16_t* my, int16_t* mz);
int selfTest(int16_t* mx, int16_t* my, int16_t* mz);
int checkNorth(int16_t* mx, int16_t* my, int16_t* mz);
//Post processing functions:
double gpsDisplacement(double *lat1, double *long1, double *lat2, double *long2);

//Global variables (passing b/w threads):
volatile double INSrecT, BARrecT, GPSrecT; //Timestamps
volatile float INSrecX, INSrecY, INSrecZ; //Accelerometer (X, Y, Z)
volatile double GPSrecN, GPSrecE; //GPS (Northing, Easting)
volatile float BARrecA, BARrecC; //Barometer (Altitude, Celsius)

volatile double ACLrawT, MAGrawT; //Timestamps
volatile float ACLrawX, ACLrawY, ACLrawZ; //Accelerometer Raw Data (X, Y, Z)
volatile float GYRrawX, GYRrawY, GYRrawZ; //Gyroscope Raw Data (X, Y, Z)
volatile float MAGrawX, MAGrawY, MAGrawZ; //Magnetometer Raw Data (X, Y, Z)

volatile int gStop;
struct tm * timeinfo;
struct timespec gettime_now;
struct timespec startTime;
time_t rawtime;
//

int main() { //create files, start INS, GPS, and post processing threads, listen for user input to end program
    {
		//Set up GPIO
		wiringPiSetupGpio(); //set up using Broadcom GPIO pin numbers
		
		pinMode(11, OUTPUT); //set LED pin to output
		pinMode(7, INPUT); //set PB pin to input
		pullUpDnControl(7, PUD_UP);
		digitalWrite(11, 1); //turn on LED on to signal start of program to user
		sleep(2);
		digitalWrite(11, 0); //
		usleep(200000);
		while(digitalRead(7)){ // flash to inform user that program has started
			digitalWrite(11, 1);
			usleep(100000);
			digitalWrite(11, 0);
			usleep(100000);
		}
		while(!digitalRead(7)){ usleep(10000); } //wait for user to let go of button
		usleep(250000); //wait .25 seconds for de-bouncing
		
		//File creation
        // Get current time with which to timestamp filenames:
        time ( &rawtime );
        timeinfo = localtime ( &rawtime );
		
        char str[60];
		
        // Create PROC file:
        pthread_mutex_lock(&procfile_mutex);
        strftime(str, 60, "/home/pi/SP14/Data/PROC_%F_%H-%M-%S.txt", timeinfo);
        // procFile = fopen(str,"w");
        procFile.open(str);
        
        if(procFile == 0){
            cout << "PROC file creation failed!" << endl;
            return 0;
        } else {
            cout << "File " << str << " opened successfully!" << endl;
        }
        pthread_mutex_unlock(&procfile_mutex);
				
        // Create PROCraw file:
        pthread_mutex_lock(&procrawfile_mutex);
        strftime(str, 60, "/home/pi/SP14/DataRaw/PROCraw_%F_%H-%M-%S.txt", timeinfo);
        // procRawFile = fopen(str,"w");
		procRawFile.open(str);
        
        if(procRawFile == 0){
            cout << "PROCraw file creation failed!" << endl;
            return 0;
        } else {
            cout << "File " << str << " opened successfully!" << endl;
        }
        pthread_mutex_unlock(&procrawfile_mutex); 
		
		// Reserve ssBuffer1/ssBuffer2 memory:
		try{
			ssBuffer1.str().reserve(resMem*1024*1024);
			cout << "String Stream Buffer 1: " << resMem << "MB Reserved successfully." << endl;
		}catch (std::bad_alloc e){
			cout << "Failed to reserve " << resMem << "MB of space." << endl;
			return(0);
		}
		
		try{
			ssBuffer2.str().reserve(resMem*1024*1024);
			cout << "String Stream Buffer 2: " << resMem << "MB Reserved successfully." << endl;
		}catch (std::bad_alloc e){
			cout << "Failed to reserve " << resMem << "MB of space." << endl;
			return(0);
		}
    }
    {	//Initialization and Main Loop
        
		//Starting conditions:		
		gStop = 1; //threads will wait while gStop > 0
		
		//Set up magnetometer to notify user of northern direction
		magInit(); //makes magnetometer separate I2C device
		int16_t mx, my, mz;
		cout 	<< "Magnetometer self test..." 
				<< mx << ", " << my << ", " << mz << endl;
		if(!selfTest(&mx, &my, &mz))
			cout 	<< "Magnetometer self test failed! Readings (X, Y, Z): " 
					<< mx << ", " << my << ", " << mz << endl;
		
		//Wait for user to push button (ideally when facing North)
		cout << "Orient Device North, Press Button to Start (Press again to stop)..." << endl;
		while(digitalRead(7)){
			getMag(&mx, &my, &mz);
			if(checkNorth(&mx, &my, &mz)){
				digitalWrite(11, 1); //turn on LED on
				cout << "N" << flush;
			} else {
				digitalWrite(11, 0); //turn on LED off at start
				cout << "." << flush;
			}
			usleep(150000);
		} //wait for user to press button
		
		while(!digitalRead(7)){ usleep(10000); } //wait for user to let go of button
		usleep(250000); //wait .25 seconds for de-bouncing
		
		digitalWrite(11, 0); //turn LED off
		cout << endl << endl;
		
		//Reset Global Start Time:
		clock_gettime(CLOCK_REALTIME, &gettime_now);
		startTime = gettime_now;
		
        gStop = -2; //threads will run while gStop < 0
        //Start threads
        int rc;
        rc = pthread_create(&insThread, 0, thread_INS, 0);
        if (rc){
            cout<< "ERROR; return code from pthread_create(ins) is " << rc << endl;
            exit(-1);
        }
        rc = pthread_create(&gpsThread, 0, thread_GPS, 0);
        if (rc){
           cout<< "ERROR; return code from pthread_create(gps) is " << rc << endl;
            exit(-1);
        }
        rc = pthread_create(&barThread, 0, thread_BAR, 0);
        if (rc){
           cout<< "ERROR; return code from pthread_create(bar) is " << rc << endl;
            exit(-1);
        }
		
		//Wait 25 seconds to start PROCthread (don't record data until gyro is stable)
		for(int i = 0; i < 250 && digitalRead(7); i++){ // flash to inform user that program has started
			digitalWrite(11, 1);
			usleep(49000);
			digitalWrite(11, 0);
			usleep(49000);
		}		
		while(!digitalRead(7)){ usleep(10000); } //wait for user to let go of button
		usleep(250000); //wait .25 seconds for de-bouncing
		
        gStop = -1; //threads will run while gStop < 0
		
		//Reset Global Start Time:
		clock_gettime(CLOCK_REALTIME, &gettime_now);
		startTime = gettime_now;
		
        rc = pthread_create(&procThread, 0, thread_PROC, 0);
        if (rc){
            cout<< "ERROR; return code from pthread_create(proc) is " << rc << endl;
            exit(-1);
        }
		
        //initiate curses (ideally to clear the screen)
        initscr();
        timeout(0);
        intrflush(stdscr, FALSE);
        cout << endl << "Running" << flush;
        char key = ' ';
		int t = 0;
        while(key != 'e' && digitalRead(7)){ //check for button press
            //threads are running
            
            //grabbing key to stop operation (former INS code):
            key = getch();
            
			usleep(499000);
			if(t%2 == 0){
				cout << " (" << (t/2)/60 << " m, " << (t/2)%60 << " s) " << flush;
				refresh();
				digitalWrite(11, 0); //turn LED off
			}else if(t%10000 == 0 && t > 1){ //~50 MB after 5200 seconds, so offload at 5k seconds (~1.44 hrs)
				procFile << ssBuffer1.str(); //write stringStream to file
				ssBuffer1.str(std::string());
				procRawFile << ssBuffer2.str(); //write stringStream to file
				ssBuffer2.str(std::string());
			}else{
				clear();
				refresh();
				digitalWrite(11, 1); //turn LED on
			}
			t++;
        }
        endwin();
		digitalWrite(11, 0); //turn LED off
        
/*** EXIT THREADS ***/
        gStop = 1; //threads will stop when !(gStop < 0)
        while(gStop < NUMTHREADS){}
/*** WAIT FOR INSFILE MUTEX LOCK ***/
        pthread_mutex_lock(&insfile_mutex);
/*** WAIT FOR GPSFILE MUTEX LOCK ***/
        pthread_mutex_lock(&gpsfile_mutex);
/*** WAIT FOR BARFILE MUTEX LOCK ***/
        pthread_mutex_lock(&barfile_mutex);
/*** WAIT FOR PROCFILE MUTEX LOCK ***/
        pthread_mutex_lock(&procfile_mutex);
/*** WAIT FOR PROCRAWFILE MUTEX LOCK ***/
        pthread_mutex_lock(&procrawfile_mutex);
/*** CLOSE ALL FILES ***/
		procFile << ssBuffer1.str(); //write stringStream to file
       	ssBuffer1.str(std::string());
		
		procRawFile << ssBuffer2.str(); //write stringStream to file
       	ssBuffer2.str(std::string());
		
		procFile.close();
		procRawFile.close();
		
        // fclose(procFile);
        // fclose(procRawFile);
/*** RELEASE/DELETE ALL MUTEXES ***/
        pthread_mutex_destroy(&insfile_mutex);
        pthread_mutex_destroy(&gpsfile_mutex);
        pthread_mutex_destroy(&barfile_mutex);
        pthread_mutex_destroy(&procfile_mutex);
        pthread_mutex_destroy(&procrawfile_mutex);
    }
	
	
    return(0);
}

// INS THREAD ///////////////////////////////////////////////////////////////////////////
void *thread_INS(void*){
    //Variable instantiation
    // MPU control/status vars
    MPU6050 mpu;
	char sensitivity = 1; //sensitivity: 0 = +/- 2g, 1 = +/- 4g, 2 = +/- 8g, 3 = +/- 16g
    //uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
    uint8_t devStatus;   // return status after each device operation (0 = success, !0 = error)
    uint16_t packetSize; // expected DMP packet size (default is 42 bytes)
    uint16_t fifoCount; // count of all bytes currently in FIFO
    uint8_t fifoBuffer[64]; // FIFO storage buffer

	// For Raw Data:
	int16_t ax, ay, az;
	int16_t gx, gy, gz;
	//int16_t mx, my, mz;

    // Orientation/Motion vars - used by INS_thread
    Quaternion q;        // [w, x, y, z]       quaternion container
    VectorInt16 aa;      // [x, y, z]          accel sensor measurements
    VectorInt16 aaReal;  // [x, y, z]          gravity-free accel sensor measurements
    VectorInt16 aaWorld; // [x, y, z]         world-frame accel sensor measurements
    VectorFloat gravity; // [x, y, z]         gravity vector
    //float euler[3];      // [psi, theta, phi]  Euler angle container
    float ypr[3];        // [yaw, pitch, roll] yaw/pitch/roll container and gravity vector
    
    { //Initialization
        // Initialize the DMP (talks with accel/gyro):
        // initialize device
        //cout << "Initializing I2C devices..." << endl;
        //insFile << "Initializing I2C devices..." << endl;
        mpu.initialize(sensitivity);
        // verify connection
        //cout << "Testing device connections..." << endl;
        //insFile << "Testing device connections..." << endl;
        if(mpu.testConnection()){
              //cout << "MPU6050 connection successful" << endl;
              //insFile << "MPU6050 connection successful" << endl;
        }else{
            //cout << "MPU6050 connection failed" << endl;
            //insFile << "MPU6050 connection failed" << endl;
            //insFile.close();
			//pthread_exit(0);
        }
        // load and configure the DMP
        //cout << "Initializing DMP..." << endl;
        //insFile << "Initializing DMP..." << endl;
        devStatus = mpu.dmpInitialize();
        // make sure it worked (returns 0 if so)
        if (devStatus == 0) {
            // turn on the DMP, now that it's ready
            //cout << "\nEnabling DMP..." << endl;
            //insFile << "\r\nEnabling DMP..." << endl;
            mpu.setDMPEnabled(true);
            //mpuIntStatus = mpu.getIntStatus();
			
            // set our DMP Ready flag so the main loop() function knows it's okay to use it
            //cout << "DMP ready!" << endl;
            //insFile << "DMP ready!" << endl;
            // get expected DMP packet size for later comparison
            packetSize = mpu.dmpGetFIFOPacketSize();
        } else {
            // ERROR!
            // 1 = initial memory load failed
            // 2 = DMP configuration updates failed
            // (if it's going to break, usually the code will be 1)
            //cout << "DMP Initialization failed (code " << devStatus << ")" << endl;
            //insFile << "DMP Initialization failed (code " << devStatus << ")" << endl;
            //insFile.close();
			//pthread_exit(0);
       }
        ////////////////////////////////////////////////////////////////////
    }
	while (gStop > 0){ usleep(1000); } //wait to start
    int i = 0;
	double curTime = 0.0; //seconds since start
	double curTimeR = 0.0;
    while(gStop < 0){ //Main loop
        ////////////////////////////////// DATA ACQUISITION:
        // get current FIFO count
        fifoCount = mpu.getFIFOCount();
		if(gStop > -2){
			clock_gettime(CLOCK_REALTIME, &gettime_now);
			curTime = 
				(gettime_now.tv_sec - startTime.tv_sec) +
				((double)(gettime_now.tv_nsec - startTime.tv_nsec)/(double)1E9);
        }
		if (fifoCount == 1024) {
            // reset so we can continue cleanly
            mpu.resetFIFO();
            cout << "FIFO overflow!" << flush;
            // otherwise, check for DMP data ready interrupt (this should happen frequently)
            fifoCount = mpu.getFIFOCount();
        } if (fifoCount >= 42) {
            //if(fifoCount > 42) //used as a sort of warning message. Removed due to clutter. Overflow occurs at 1024 bytes.
            //    cout << "!!! fifoCount = " << fifoCount << endl;
            // read a packet from FIFO
            mpu.getFIFOBytes(fifoBuffer, packetSize);
			// Process:
            mpu.dmpGetQuaternion(&q, fifoBuffer);
            mpu.dmpGetAccel(&aa, fifoBuffer);
            mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
            mpu.dmpGetGravity(&gravity, &q);
            /* Leave commented out - this was the old way of adjusting for gravity:
			 * gravity.x *= grav_Factor;
             * gravity.y *= grav_Factor;
             * gravity.z *= grav_Factor;
			 */
            //adjust for gravity based on sensitivity (±2g, others)
            mpu.dmpGetLinearAccel(&aaReal, &aa, &gravity, sensitivity);
            mpu.dmpGetLinearAccelInWorld(&aaWorld, &aaReal, &q);
			mpu.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);
            clock_gettime(CLOCK_REALTIME, &gettime_now);
			curTimeR = 
				(gettime_now.tv_sec - startTime.tv_sec) +
				((double)(gettime_now.tv_nsec - startTime.tv_nsec)/(double)1E9);
            ////////////////////////////////// DATA OUTPUT/RECORDING:
/*** WAIT FOR INSFILE MUTEX LOCK ***/
            // pthread_mutex_lock(&insfile_mutex);
			
			//save data to memory for PROC.txt:
			INSrecT = curTime; //timestamp
			INSrecX = aaWorld.x;
			INSrecY = aaWorld.y;
			INSrecZ = aaWorld.z;
			
			//save data to memory for PROCraw.txt:
			ACLrawT = curTimeR; //timestamp
			// ACLrawX = ax;
			// ACLrawY = ay;
			// ACLrawZ = az;
			ACLrawX = aa.x;
			ACLrawY = aa.y;
			ACLrawZ = aa.z;
			GYRrawX = gx;
			GYRrawY = gy;
			GYRrawZ = gz;
			
            ////insFile << "Record: " << i << " at: " << curTime << "-s: /t" <<
			//	"X:" << INSrecX << " Y: " << INSrecY << " Z: " << INSrecZ << endl;
			
            i++;
/*** RELEASE INSFILE MUTEX LOCK ***/
            // pthread_mutex_unlock(&insfile_mutex);
            //release mutex lock before next line, allowing thread_PROC to read INSdata.txt
        }
		usleep(1000);	//~3ms is the maximum without causing FIFO overflow as of 10/22/13
						//as of 3/31/14, 3ms causes overflow (with multithreading). Adjusted down to 1ms. Repositioned from inside (fifoCount > 42){ }
    }
    gStop++;
    pthread_exit(0);
}

// GPS THREAD ///////////////////////////////////////////////////////////////////////////
void *thread_GPS(void*){
    //Variable instantiation
    //Serial and GPS variables
    char gpsData; //number of GPS bytes available
	char gpsMessage[100]; //serial string read in from GPS unit
	char gpgga[] = "$GPGGA";
	
    unsigned char arPos; //array position char
    double easting;
    double northing;
		    
    { //GPS data acquisition
        /*Initialize serial*/
        /*Note: has to be done in order! Cannot access GPS without initializing serial port!*/

        bool ok = false;
        
		// GPS connection handle
		char serialPort[] = "/dev/ttyAMA0";
        char handle = serialOpen(serialPort, 9600);
        // Serial port opened.
        
		while (gStop > 0){ usleep(1000); } //wait to start
		
        double curTime = 0.0;
        while(gStop < 0){ //infinite loop, each iteration fetches/writes a new GPS coordinate to file
            ok = false;
            do{
				gpsData = serialDataAvail(handle);
                if(gpsData > 0){
                    ok = true;
                    arPos = 0;
					do{
                        gpsMessage[arPos] = serialGetchar(handle);
                        arPos++;
                    }while(gpsMessage[arPos-1] != 10 && arPos < 100);
                    //ASCII 10 is the line feed (LF) character
                    for(int i = 0; i < 6; i++){
                        if(gpsMessage[i] != gpgga[i])
                            ok = false;
                    }
                }
            }while(!ok && gStop < 0);
            string gpggaMessage(gpsMessage);
            
            easting = strtold(gpggaMessage.substr(30,3).c_str(), NULL) +
                (strtold(gpggaMessage.substr(33,7).c_str(), NULL)/60.0000D);
            northing = strtold(gpggaMessage.substr(18,2).c_str(), NULL) +
                (strtold(gpggaMessage.substr(20,7).c_str(), NULL)/60.0000D);
			
            if(gpggaMessage[41] != 'E')
                easting *= -1;
            if(gpggaMessage[28] != 'N')
                northing *= -1;
			
			if(gStop > -2){
				clock_gettime(CLOCK_REALTIME, &gettime_now);
				curTime = 
					(gettime_now.tv_sec - startTime.tv_sec) +
					((double)(gettime_now.tv_nsec - startTime.tv_nsec)/(double)1E9);
			}
/*** WAIT FOR GPSFILE MUTEX LOCK ***/
            // pthread_mutex_lock(&gpsfile_mutex);
            
			GPSrecT = curTime; //timestamp
			GPSrecN = northing;
			GPSrecE = easting;
/*** RELEASE GPSFILE MUTEX LOCK ***/
            // pthread_mutex_unlock(&gpsfile_mutex);
            usleep(999000); //~1s delay
        }
        serialClose(handle);
        /*END SERIAL CODE*/
    }
    gStop++;
    pthread_exit(0);
}

// BAROMETER THREAD ///////////////////////////////////////////////////////////////////////////
void *thread_BAR(void*){
	Adafruit_BMP085_Unified bmp180;
	int errorCode = bmp180.begin();
	if(errorCode < 0){
		//barFile << "Unable to connect to the BMP180! Reason: " << errorCode << endl;
		pthread_exit(0);
	}
	
	float pressure = 0;
	float temp = 0;
	float seaLevel = 1021; //This should be a local value. From website in 3/25/14 documentation (hpa)
	float altitude = 0;
	
    { //Barometer data acquisition
        
		while (gStop > 0){ usleep(1000); } //wait to start
        int i=0;
        double curTime = 0.0;
        while(gStop < 0){ //infinite loop, writes a new pressure/magnetometer reading to mem
            //get pressure
            bmp180.getPressure(&pressure);
            bmp180.getTemperature(&temp);
            altitude = bmp180.pressureToAltitude(seaLevel, pressure/100.0, temp); //pressure in hPa required (1 hPa = 100 Pa)
            
            if(gStop > -2){
				clock_gettime(CLOCK_REALTIME, &gettime_now);
				curTime = 
					(gettime_now.tv_sec - startTime.tv_sec) +
					((double)(gettime_now.tv_nsec - startTime.tv_nsec)/(double)1E9);
			}
				
/*** WAIT FOR BARFILE MUTEX LOCK ***/
            // pthread_mutex_lock(&barfile_mutex);
            //update BAR file
            // //barFile << "Record: " << i << " at " << curTime << " seconds. Pressure: " << pressure << " Temp: " << temp << " Altitude: " << altitude << endl;
            i++;
			BARrecT = curTime; //timestamp
			BARrecA = altitude;
			BARrecC = temp;
/*** RELEASE BARFILE MUTEX LOCK ***/
            // pthread_mutex_unlock(&barfile_mutex);
            usleep(998000); //~1s delay
        }
    }
    gStop++;
    pthread_exit(0);
}

// PROC THREAD //////////////////////////////////////////////////////////////////////////
void *thread_PROC(void*){
    //Variable instantiation
    //Kalman filter variables - used by PROC_thread
    pthread_mutex_lock(&procfile_mutex);
    pthread_mutex_lock(&procrawfile_mutex);
    { //Post-processing loop
        //wait for mutex lock to read any files
		
		while (gStop > 0){ usleep(1000); } //wait to start
        //post-processing variables
		double INSprevT = -1.0;
        double GPSprevT = -1.0;
        double BARprevT = -1.0;
		
		double INSdeltaT, GPSdeltaT, BARdeltaT;
		double prevLat, prevLong, curLat, curLong; //Lat == Northing, Long = Easting
		double GPSdispX = 0, GPSdispY = 0, GPSvelX = 0, GPSvelY = 0;
        while(gStop < 0){			
			// pthread_mutex_lock(&insfile_mutex);
			INSdeltaT = INSrecT - INSprevT;
            if(INSdeltaT > 0){
				if(INSprevT != -1.0){
					//Process raw acceleration data
					double aWNorm[3]; // This is "N, E, Up" - "+Y, -X, +Z" from accelerometer
					aWNorm[0] = ((double)INSrecX/(double)0x7FFF)*4*9.81;
					aWNorm[1] = ((double)INSrecY/(double)0x7FFF)*4*9.81;
					aWNorm[2] = ((double)INSrecZ/(double)0x7FFF)*4*9.81;
					//Save raw acceleration data to file
					ssBuffer1 << "                   INS: @" << setprecision(12) << INSrecT << " s, " << 
						"N: " << aWNorm[0] << " E: " << aWNorm[1] << " Up: " << aWNorm[2] << endl;
					//fprintf(procFile, "                   INS: @%.12f s, N: %.12f E: %.12f Up: %.12f\r\n", INSrecT, aWNorm[0], aWNorm[1], aWNorm[2]);
					ssBuffer2 << "@" << setprecision(8) << ACLrawT << " s," 
						<< " aX: " << ACLrawX << " aY: " << ACLrawY << " aZ: " << ACLrawZ 
						<< " gX: " << GYRrawX << " gY: " << GYRrawY << " gZ: " << GYRrawZ << endl;
					//fprintf(procRawFile, "@%.12f s, Accel: X: %.12f Y: %.12f Z: %.12f Gyro: X: %.12f Y: %.12f Z: %.12f\r\n", ACLrawT, ACLrawX, ACLrawY, ACLrawZ, GYRrawX, GYRrawY, GYRrawZ);
				}
				INSprevT = INSrecT;
            }
			
            // pthread_mutex_unlock(&insfile_mutex);
			
            // pthread_mutex_lock(&gpsfile_mutex);
			GPSdeltaT = GPSrecT - GPSprevT;
            if(GPSdeltaT > 0){
				//Process raw GPS data
				curLat = GPSrecN;
				curLong = GPSrecE;
				if(GPSprevT == -1.0){ //if first non-zero data point
					prevLat = GPSrecN;
					prevLong = GPSrecE;
				} else {
					//Calculate X and Y displacement (meters):
					GPSdispX = gpsDisplacement(&prevLat, &prevLong, &prevLat, &curLong); //East-West displacement. Use same latitude/northing.
					if(curLong < prevLong)
						GPSdispX *= -1;
					GPSdispY = gpsDisplacement(&prevLat, &prevLong, &curLat, &prevLong); //North-South displacement. Use same longitude/easting.
					if(curLat < prevLat)
						GPSdispY *= -1;
					GPSvelX = GPSdispX / GPSdeltaT;
					GPSvelY = GPSdispY / GPSdeltaT;
					//update previous lat/long values:
					prevLat = curLat;
					prevLong = curLong;
					//Save raw GPS data to file
					ssBuffer1 << "*******************GPS: @" << setprecision(12) << GPSrecT << " s, " << 
						"N: " << curLat << " E: " << curLong << 
						" dispX: " << GPSdispX << " dispY: " << GPSdispY << 
						" velX: " << GPSvelX << " velY: " << GPSvelY << endl;
					//fprintf(procFile, "*******************GPS: @%.12f s, N: %.12f E: %.12f dispX: %.12f dispY: %.12f velX: %.12f velY: %.12f\r\n", GPSrecT, curLat, curLong, GPSdispX, GPSdispY, GPSvelX, GPSvelY);
				}
				GPSprevT = GPSrecT;
            }
            // pthread_mutex_unlock(&gpsfile_mutex);
			
            // pthread_mutex_lock(&barfile_mutex);
			BARdeltaT = BARrecT - BARprevT;
            if(BARdeltaT > 0){
				if(BARprevT != -1.0){
					//Process raw altitude data
					
					//Save raw altitude data to file
					ssBuffer1 << "**               **BAR: @" << setprecision(12) << BARrecT << " s, " << 
						"A: " << setprecision(7) << BARrecA << " T: " << BARrecC << endl;
					//fprintf(procFile, "**               **BAR: @%.12f s, A: %f T: %f\r\n", BARrecT, BARrecA, BARrecC);
				}
				BARprevT = BARrecT;
            }
            // pthread_mutex_unlock(&barfile_mutex);
			
			usleep(100); //delay to lower CPU utilization
        }
    }
    pthread_mutex_unlock(&procfile_mutex);
    pthread_mutex_unlock(&procrawfile_mutex);
    gStop++;
    pthread_exit(0);
}

void magInit(){
    MPU6050 mpu;
	mpu.initialize();
	I2Cdev::writeByte(0x68, 0x37, 0x02); //set i2c bypass enable pin to true to access magnetometer
}

void getMag(int16_t* mx, int16_t* my, int16_t* mz) {
	// Set single measurement mode 
	I2Cdev::writeByte(0x0C, 0x0A, 0x01); //write 0x01 to CNTL register
	// Check Data Ready or not by any of the following method.
	uint8_t DRDY = 0;
	while(!DRDY){
		I2Cdev::readByte(0x0C, 0x02, &DRDY);
	}
	readMagData(mx, my, mz);
}

void readMagData(int16_t* mx, int16_t* my, int16_t* mz){
	uint8_t buffer[6];
	I2Cdev::readBytes(0x0C, 0x03, 6, buffer);
	*mx = (((int16_t)buffer[1]) << 8) | buffer[0];
    *my = (((int16_t)buffer[3]) << 8) | buffer[2];
    *mz = (((int16_t)buffer[5]) << 8) | buffer[4];
}

int selfTest(int16_t* mx, int16_t* my, int16_t* mz) {
	// (1) Set Power-down mode 
	I2Cdev::writeByte(0x0C, 0x0A, 0x00); //write 0x00 to CNTL register
	// (2) Write “1” to SELF bit of ASTC register
	I2Cdev::writeByte(0x0C, 0x0C, 0x40);
	// (3) Set Self-test Mode 
	I2Cdev::writeByte(0x0C, 0x0A, 0x08);
	// (4) Check Data Ready or not by any of the following method.
	uint8_t DRDY = 0;
	while(!DRDY){
		I2Cdev::readByte(0x0C, 0x02, &DRDY);
	}
	//  When Data Ready, proceed to the next step. 
	// (5) Read measurement data (HXL to HZH)
	readMagData(mx, my, mz);
	// (6) Write “0” to SELF bit of ASTC register
	I2Cdev::writeByte(0x0C, 0x0C, 0x00);
	
	//check self-test pass criteria (return 0 if fail):
	if(	*mx < -100  || *mx >  100 || 
		*my < -100  || *my >  100 || 
		*mz < -1000 || *mz > -300)
		return 0;
	else
		return 1;
}

int checkNorth(int16_t* mx, int16_t* my, int16_t* mz) {
	float magnitude, normX, normY, normZ;
	magnitude = sqrt(pow(*mx, 2) + pow(*my, 2) + pow(*mz, 2));
	normX = *mx / magnitude;
	normY = *my / magnitude;
	normZ = *mz / magnitude;
	if(	normX < 0.70 && normX > 0.55 &&
		normY < 0.85 && normY > 0.70 &&
		normZ < 0.10 && normZ > -0.10)
		return 1;
	else
		return 0;
}

double gpsDisplacement(double *lat1, double *long1, double *lat2, double *long2){
	
	double PI = 4.0*atan(1.0);
	
	double dlat1=*lat1*(PI/180);
	double dlong1=*long1*(PI/180);
	double dlat2=*lat2*(PI/180);
	double dlong2=*long2*(PI/180);

	double dLong=dlong1-dlong2;
	double dLat=dlat1-dlat2;

	double aHarv= pow(sin(dLat/2.0),2.0)+cos(dlat1)*cos(dlat2)*pow(sin(dLong/2),2);
	double cHarv=2*atan2(sqrt(aHarv),sqrt(1.0-aHarv));
	
	const double earth=6371009; //mean Earth radius (meters)
	double distance=earth*cHarv;
	
	return distance;
}